/*
Module : WTimer.h
Purpose: Defines the interface for an C++ wrapper class for waitable timers
Created: PJN / 06-08-2000
History: PJN / 08-11-2002 1. Removed some unnecessary comments in the module
                          2. Updated copyright details in the module
         PJN / 08-02-2004 1. Completion routine can now be passed an LPVOID parameter just like 
                          the SDK equivalent. Thanks to Erenst F. Wepener for suggesting this 
                          addition.
                          2. Minor updates to the code to allow the code to compile on VC 5 
                          without the requirement for the Platform SDK to be installed.
         PJN / 26-06-2006 1. Updated copyright details.
                          2. Removed various defines in code which were taken from the Platform SDK. 
                          This does mean that the code now requires the Platform SDK if compiled using VC 6.
                          3. Optimized CWaitableTimer constructor code.
                          4. Combined the functionality of the _WAITABLE_TIMER_DATA class into the main 
                          CWaitableTimer class.
                          5. Code now uses new C++ style casts rather than old style C casts where necessary. 
                          6. Reviewed all TRACE statements for correctness
         PJN / 01-06-2008 1. Code now compiles cleanly using Code Analysis (/analyze)
                          2. The code has now been updated to support VC 2005 or later only.  
                          3. Updated the sample app to clean compile on VC 2008
         PJN / 16-02-2014 1. Updated copyright details
                          2. Made the class into a Header only implementation. The WTimer.cpp module is now
                          defunct.
                          3. Code now no longer uses GetProcAddress to obtain the waitable timer functions
                          since waitable timers have been available since Windows 2000.
                          4. Made the code independent of MFC.
                          5. Updated the code to clean compile using /analyze
                          6. Updated the code to clean compile on VC 2010, 2012 & 2013
         PJN / 04-12-2015 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2015.
                          3. All the class methods have had SAL annotations added 
                          4. The SetOnceOffAbsolute, SetOnceOffRelative, SetPeriodicAbsolute & 
                          SetPeriodicRelative methods has been reworked to expose all the parameters of 
                          the underlying SetWaitableTimer API.
                          5. Removed unnecessary ATLTRACE calls from various methods
                          6. Added support for CreateWaitableTimerEx and SetWaitableTimerEx APIs to class

Copyright (c) 2000 - 2015 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __WTIMER_H__
#define __WTIMER_H__

#ifndef CWAITABLETIMER_EXT_CLASS
#define CWAITABLETIMER_EXT_CLASS
#endif

#ifndef _In_
#define _In_
#endif //#ifndef _In_

#ifndef _In_opt_
#define _In_opt_
#endif //#ifndef _In_opt_



/////////////////////////////// Classes ///////////////////////////////////////

class CWAITABLETIMER_EXT_CLASS CWaitableTimer
{
public:
//Constructors / Destructors
  CWaitableTimer() : m_hTimer(NULL)
  {
  }
  
  virtual ~CWaitableTimer()
  {
    Close();
  }

//Methods
  BOOL Open(_In_ LPCTSTR pszTimerName, _In_ BOOL bInherit = FALSE, _In_ DWORD dwAccess = TIMER_ALL_ACCESS)
  {
    //Free up any existent timer
    Close();

    //Call the SDK function
    m_hTimer = OpenWaitableTimer(dwAccess, bInherit, pszTimerName);
    return (m_hTimer != NULL);
  }
  
  BOOL Create(_In_ BOOL bManualReset, _In_opt_ LPCTSTR pszTimerName = NULL, _In_opt_ LPSECURITY_ATTRIBUTES lpTimerAttributes = NULL)
  {
    //Free up any existent timer
    Close();

    //Call the SDK function
    m_hTimer = CreateWaitableTimer(lpTimerAttributes, bManualReset, pszTimerName);
    return (m_hTimer != NULL);
  }

#if (_WIN32_WINNT >= 0x0600)
__if_exists(CreateWaitableTimerEx)
{
  BOOL Create(_In_ DWORD dwFlags, _In_ DWORD dwDesiredAccess, _In_opt_ LPCTSTR lpTimerName = NULL, _In_opt_ LPSECURITY_ATTRIBUTES lpTimerAttributes = NULL)
  {
    //Free up any existent timer
    Close();

    //Call the SDK function
    m_hTimer = CreateWaitableTimerEx(lpTimerAttributes, lpTimerName, dwFlags, dwDesiredAccess);
    return (m_hTimer != NULL);
  }
} //__if_exists(CreateWaitableTimerEx)
#endif //#if (_WIN32_WINNT >= 0x0600)
  
  void Close()
  {
    if (m_hTimer != NULL)
    {
      //Free up any existent timer
      CloseHandle(m_hTimer);
      m_hTimer = NULL;
    }
  }
  
  operator HANDLE() 
  { 
    return m_hTimer; 
  }
  
  BOOL Attach(_In_opt_ HANDLE hTimer)
  {
    Close();
    m_hTimer = hTimer;
    return TRUE;
  }
  
  HANDLE Detach()
  {
    HANDLE hReturn = m_hTimer;
    m_hTimer = NULL;
    return hReturn;
  }

  BOOL SetOnceOffAbsolute(_In_ const SYSTEMTIME& DueTime, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine = NULL, _In_opt_ LPVOID lpArgToCompletionRoutine = NULL, _In_ BOOL fResume = FALSE)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the SystemTime to the filetime struct
    FILETIME ft;
    if (!SystemTimeToFileTime(&DueTime, &ft))
      return FALSE;

    LARGE_INTEGER li;
    li.LowPart = ft.dwLowDateTime;
    li.HighPart = ft.dwHighDateTime; 
    return SetWaitableTimer(m_hTimer, &li, 0, pfnCompletionRoutine, lpArgToCompletionRoutine, fResume);
  }

  BOOL SetOnceOffRelative(_In_ DWORD dwMilliseconds, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine = NULL, _In_opt_ LPVOID lpArgToCompletionRoutine = NULL, _In_ BOOL fResume = FALSE)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the milliseconds to FILETIME representation (100-nanoseconds)
    //We also need to make it negative
    __int64 nFileTimes = static_cast<__int64>(-10000) * static_cast<__int64>(dwMilliseconds);
    LARGE_INTEGER li;
    li.LowPart = static_cast<DWORD>(nFileTimes & 0xFFFFFFFF);
    li.HighPart = static_cast<LONG>(nFileTimes >> 32); 
    return SetWaitableTimer(m_hTimer, &li, 0, pfnCompletionRoutine, lpArgToCompletionRoutine, fResume);
  }
  
  BOOL SetPeriodicAbsolute(_In_ const SYSTEMTIME& DueTime, _In_ LONG lPeriod, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine = NULL, _In_opt_ LPVOID lpArgToCompletionRoutine = NULL, _In_ BOOL fResume = FALSE)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the SystemTime to the filetime struct
    FILETIME ft;
    if (!SystemTimeToFileTime(&DueTime, &ft))
      return FALSE;

    LARGE_INTEGER li;
    li.LowPart = ft.dwLowDateTime;
    li.HighPart = ft.dwHighDateTime; 
    return SetWaitableTimer(m_hTimer, &li, lPeriod, pfnCompletionRoutine, lpArgToCompletionRoutine, fResume);
  }
  
  BOOL SetPeriodicRelative(_In_ DWORD dwMilliseconds, _In_ LONG lPeriod, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine = NULL, _In_opt_ LPVOID lpArgToCompletionRoutine = NULL, _In_ BOOL fResume = FALSE)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the milliseconds to FILETIME representation (100-nanoseconds)
    //We also need to make it negative
    __int64 nFileTimes = static_cast<__int64>(-10000) * static_cast<__int64>(dwMilliseconds);
    LARGE_INTEGER li;
    li.LowPart = static_cast<DWORD>(nFileTimes & 0xFFFFFFFF);
    li.HighPart = static_cast<LONG>(nFileTimes >> 32); 
    return SetWaitableTimer(m_hTimer, &li, lPeriod, pfnCompletionRoutine, lpArgToCompletionRoutine, fResume);
  }

#if (_WIN32_WINNT >= _WIN32_WINNT_WIN7)
__if_exists(PREASON_CONTEXT)
{
  BOOL SetOnceOffAbsolute(_In_ const SYSTEMTIME& DueTime, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine, _In_opt_ LPVOID lpArgToCompletionRoutine, _In_opt_ PREASON_CONTEXT WakeContext, _In_ ULONG TolerableDelay)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the SystemTime to the filetime struct
    FILETIME ft;
    if (!SystemTimeToFileTime(&DueTime, &ft))
      return FALSE;

    LARGE_INTEGER li;
    li.LowPart = ft.dwLowDateTime;
    li.HighPart = ft.dwHighDateTime;
    return SetWaitableTimerEx(m_hTimer, &li, 0, pfnCompletionRoutine, lpArgToCompletionRoutine, WakeContext, TolerableDelay);
  }

  BOOL SetOnceOffRelative(_In_ DWORD dwMilliseconds, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine, _In_opt_ LPVOID lpArgToCompletionRoutine, _In_opt_ PREASON_CONTEXT WakeContext, _In_ ULONG TolerableDelay)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the milliseconds to FILETIME representation (100-nanoseconds)
    //We also need to make it negative
    __int64 nFileTimes = static_cast<__int64>(-10000) * static_cast<__int64>(dwMilliseconds);
    LARGE_INTEGER li;
    li.LowPart = static_cast<DWORD>(nFileTimes & 0xFFFFFFFF);
    li.HighPart = static_cast<LONG>(nFileTimes >> 32);
    return SetWaitableTimerEx(m_hTimer, &li, 0, pfnCompletionRoutine, lpArgToCompletionRoutine, WakeContext, TolerableDelay);
  }

  BOOL SetPeriodicAbsolute(_In_ const SYSTEMTIME& DueTime, _In_ LONG lPeriod, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine, _In_opt_ LPVOID lpArgToCompletionRoutine, _In_opt_ PREASON_CONTEXT WakeContext, _In_ ULONG TolerableDelay)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the SystemTime to the filetime struct
    FILETIME ft;
    if (!SystemTimeToFileTime(&DueTime, &ft))
      return FALSE;

    LARGE_INTEGER li;
    li.LowPart = ft.dwLowDateTime;
    li.HighPart = ft.dwHighDateTime;
    return SetWaitableTimerEx(m_hTimer, &li, lPeriod, pfnCompletionRoutine, lpArgToCompletionRoutine, WakeContext, TolerableDelay);
  }

  BOOL SetPeriodicRelative(_In_ DWORD dwMilliseconds, _In_ LONG lPeriod, _In_opt_ PTIMERAPCROUTINE pfnCompletionRoutine, _In_opt_ LPVOID lpArgToCompletionRoutine, _In_opt_ PREASON_CONTEXT WakeContext, _In_ ULONG TolerableDelay)
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Convert from the milliseconds to FILETIME representation (100-nanoseconds)
    //We also need to make it negative
    __int64 nFileTimes = static_cast<__int64>(-10000) * static_cast<__int64>(dwMilliseconds);
    LARGE_INTEGER li;
    li.LowPart = static_cast<DWORD>(nFileTimes & 0xFFFFFFFF);
    li.HighPart = static_cast<LONG>(nFileTimes >> 32);
    return SetWaitableTimerEx(m_hTimer, &li, lPeriod, pfnCompletionRoutine, lpArgToCompletionRoutine, WakeContext, TolerableDelay);
  }
} //__if_exists(PREASON_CONTEXT)
#endif //#if (_WIN32_WINNT >= _WIN32_WINNT_WIN7)

  BOOL Cancel()
  {
    //Validate our parameters
    ATLASSUME(m_hTimer != NULL); //Timer must be valid

    //Call the SDK function
    return CancelWaitableTimer(m_hTimer);
  }

protected:
//Member variables
  HANDLE m_hTimer;
};

#endif //__WTIMER_H__
